#import <UIKit/UIKit.h>
#include <AVFoundation/AVFoundation.h>

@interface                                QIOSApplicationDelegate : UIResponder <UIApplicationDelegate>
@property ( strong, nonatomic ) UIWindow* window;

@end
