#import "ios_app_delegate.h"

@interface QIOSApplicationDelegate ()

@end

@implementation QIOSApplicationDelegate

- (BOOL)application:(UIApplication*)application didFinishLaunchingWithOptions:(NSDictionary*)launchOptions
{
    return YES;
}
@end
